package com.qfs.pagan

import android.annotation.SuppressLint
import android.app.Activity
import android.app.AlertDialog
import android.app.NotificationChannel
import android.app.NotificationManager
import android.app.PendingIntent
import android.content.Context
import android.content.Intent
import android.content.IntentFilter
import android.content.pm.PackageManager
import android.content.res.Configuration
import android.database.Cursor
import android.media.midi.MidiDeviceInfo
import android.net.Uri
import android.os.Build
import android.os.Bundle
import android.os.VibrationEffect
import android.os.Vibrator
import android.os.VibratorManager
import android.provider.OpenableColumns
import android.util.Log
import android.view.ContextThemeWrapper
import android.view.KeyEvent
import android.view.LayoutInflater
import android.view.Menu
import android.view.MenuItem
import android.view.MotionEvent
import android.view.View
import android.view.ViewGroup
import android.view.WindowManager
import android.widget.EditText
import android.widget.LinearLayout
import android.widget.NumberPicker
import android.widget.TextView
import android.widget.Toast
import androidx.activity.OnBackPressedCallback
import androidx.activity.result.contract.ActivityResultContracts
import androidx.activity.viewModels
import androidx.appcompat.app.ActionBarDrawerToggle
import androidx.appcompat.app.AppCompatActivity
import androidx.constraintlayout.widget.ConstraintLayout
import androidx.core.app.ActivityCompat
import androidx.core.app.NotificationCompat
import androidx.core.app.NotificationManagerCompat
import androidx.core.content.ContextCompat
import androidx.core.net.toUri
import androidx.core.os.bundleOf
import androidx.documentfile.provider.DocumentFile
import androidx.drawerlayout.widget.DrawerLayout
import androidx.fragment.app.Fragment
import androidx.fragment.app.clearFragmentResult
import androidx.fragment.app.setFragmentResult
import androidx.lifecycle.ViewModel
import androidx.navigation.findNavController
import androidx.navigation.ui.AppBarConfiguration
import androidx.navigation.ui.navigateUp
import androidx.recyclerview.widget.RecyclerView
import com.google.android.material.button.MaterialButton
import com.qfs.apres.InvalidMIDIFile
import com.qfs.apres.Midi
import com.qfs.apres.MidiController
import com.qfs.apres.MidiPlayer
import com.qfs.apres.VirtualMidiInputDevice
import com.qfs.apres.VirtualMidiOutputDevice
import com.qfs.apres.event.BankSelect
import com.qfs.apres.event.ProgramChange
import com.qfs.apres.event.SongPositionPointer
import com.qfs.apres.event2.NoteOn79
import com.qfs.apres.soundfont.Riff
import com.qfs.apres.soundfont.SoundFont
import com.qfs.apres.soundfontplayer.SampleHandleManager
import com.qfs.apres.soundfontplayer.WavConverter
import com.qfs.apres.soundfontplayer.WaveGenerator
import com.qfs.pagan.ActionTracker.TrackedAction
import com.qfs.pagan.databinding.ActivityMainBinding
import com.qfs.pagan.opusmanager.OpusChannelAbstract
import com.qfs.pagan.opusmanager.OpusLayerBase
import com.qfs.pagan.opusmanager.OpusLineAbstract
import com.qfs.pagan.opusmanager.OpusManagerCursor
import java.io.BufferedOutputStream
import java.io.DataOutputStream
import java.io.File
import java.io.FileInputStream
import java.io.FileNotFoundException
import java.io.FileOutputStream
import java.time.LocalDateTime
import java.time.format.DateTimeFormatter
import kotlin.concurrent.thread
import kotlin.math.floor
import kotlin.math.max
import kotlin.math.min
import kotlin.math.roundToInt
import com.qfs.pagan.OpusLayerInterface as OpusManager

class MainActivity : AppCompatActivity() {
    companion object {
        init {
            System.loadLibrary("pagan")
        }
    }

    enum class PlaybackState {
        NotReady,
        Ready,
        Playing,
        Queued,
        Stopping
    }

    class MainViewModel: ViewModel() {
        var export_handle: WavConverter? = null
        var action_interface = ActionTracker()
        var opus_manager = OpusManager()

        fun export_wav(
            opus_manager: OpusLayerBase,
            sample_handle_manager: SampleHandleManager,
            target_output_stream: DataOutputStream,
            tmp_file: File, configuration: PaganConfiguration? = null,
            handler: WavConverter.ExporterEventHandler,
            ignore_global_effects: Boolean = false,
            ignore_channel_effects: Boolean = false,
            ignore_line_effects: Boolean = false,
        ) {
            val frame_map = PlaybackFrameMap(opus_manager, sample_handle_manager)
            frame_map.clip_same_line_release = configuration?.clip_same_line_release ?: true
            frame_map.parse_opus(
                ignore_global_effects,
                ignore_channel_effects,
                ignore_line_effects
            )

            val start_frame = frame_map.get_marked_frames()[0]

            // Prebuild the first buffer's worth of sample handles, the rest happen in the get_new_handles()
            for (i in start_frame .. start_frame + sample_handle_manager.buffer_size) {
                frame_map.check_frame(i)
            }

            this.export_handle = WavConverter(sample_handle_manager)

            this.export_handle?.export_wav(frame_map, target_output_stream, tmp_file, handler)
            this.export_handle = null
        }

        fun cancel_export() {
            val handle = this.export_handle ?: return
            handle.cancel_flagged = true
        }

        fun is_exporting(): Boolean {
            return this.export_handle != null
        }
    }

    val view_model: MainViewModel by viewModels()
    // flag to indicate that the landing page has been navigated away from for navigation management
    private var _has_seen_front_page = false
    private lateinit var _project_manager: ProjectManager
    lateinit var configuration: PaganConfiguration
    private lateinit var _config_path: String
    private var _integer_dialog_defaults = HashMap<String, Int>()
    private var _float_dialog_defaults = HashMap<String, Float>()
    var active_percussion_names = HashMap<Int, String>()

    private var _virtual_input_device = MidiPlayer()
    private lateinit var _midi_interface: MidiController
    private var _soundfont: SoundFont? = null
    internal var _soundfont_supported_instrument_names = HashMap<Pair<Int, Int>, String>()
    private var _sample_handle_manager: SampleHandleManager? = null
    private var _feedback_sample_manager: SampleHandleManager? = null
    private var _midi_playback_device: PlaybackDevice? = null
    private var _midi_feedback_dispatcher = MidiFeedbackDispatcher()

    private lateinit var _app_bar_configuration: AppBarConfiguration
    private lateinit var _binding: ActivityMainBinding
    private var _options_menu: Menu? = null
    private var _progress_bar: ConstraintLayout? = null
    var playback_state_soundfont: PlaybackState = PlaybackState.NotReady
    var playback_state_midi: PlaybackState = PlaybackState.NotReady
    private var _forced_title_text: String? = null
    private val _temporary_feedback_devices = Array<FeedbackDevice?>(4) {
        null
    }
    private var _current_feedback_device: Int = 0
    private var _blocker_scroll_y: Float? = null
    private var broadcast_receiver = PaganBroadcastReceiver()
    private var receiver_intent_filter = IntentFilter("com.qfs.pagan.CANCEL_EXPORT_WAV")

    // Notification shiz -------------------------------------------------
    var NOTIFICATION_ID = 0
    val CHANNEL_ID = "com.qfs.pagan" // TODO: Use String Resource
    private var _notification_channel: NotificationChannel? = null
    private var _active_notification: NotificationCompat.Builder? = null
    // -------------------------------------------------------------------

    private var _popup_active: Boolean = false

    class MultiExporterEventHandler(var activity: MainActivity, var total_count: Int): WavConverter.ExporterEventHandler {
        var working_y = 0
        var file_uri: Uri? = null
        var cancelled = false

        val notification_manager = NotificationManagerCompat.from(this.activity)

        override fun on_start() {
            if (this.working_y != 0) {
                return
            }

            this.activity.runOnUiThread {
                val btnExportProject = this.activity.findViewById<MaterialButton>(R.id.btnExportProject) ?: return@runOnUiThread
                btnExportProject.setIconResource(R.drawable.baseline_cancel_42)
                val llExportProgress = this.activity.findViewById<View>(R.id.llExportProgress) ?: return@runOnUiThread
                llExportProgress.visibility = View.VISIBLE
                val tvExportProgress = this.activity.findViewById<TextView>(R.id.tvExportProgress) ?: return@runOnUiThread
                tvExportProgress.text = "0%"
            }

            val builder = this.activity.get_notification() ?: return
            @SuppressLint("MissingPermission")
            if (this.activity.has_notification_permission()) {
                this.notification_manager.notify(this.activity.NOTIFICATION_ID, builder.build())
            }
        }

        override fun on_complete() {
            if (this.working_y < this.total_count - 1) {
                return
            }

            val builder = this.activity.get_notification()
            if (builder != null) {
                // NON functional ATM, Open file from notification
                val go_to_file_intent = Intent()
                go_to_file_intent.action = Intent.ACTION_VIEW
                go_to_file_intent.setDataAndType(this.file_uri!!, "*/*")

                val pending_go_to_intent = PendingIntent.getActivity(
                    this.activity,
                    0,
                    go_to_file_intent,
                    PendingIntent.FLAG_IMMUTABLE
                )

                builder.setContentText(this.activity.getString(R.string.export_wav_notification_complete))
                    .clearActions()
                    .setAutoCancel(true)
                    .setProgress(0, 0, false)
                    .setTimeoutAfter(5000)
                    .setSilent(false)
                    .setContentIntent(pending_go_to_intent)

                @SuppressLint("MissingPermission")
                if (this.activity.has_notification_permission()) {
                    this.notification_manager.notify(this.activity.NOTIFICATION_ID, builder.build())
                }
            }

            this.activity.feedback_msg(this.activity.getString(R.string.export_wav_feedback_complete))

            this.activity.runOnUiThread {
                val llExportProgress = this.activity.findViewById<View>(R.id.llExportProgress) ?: return@runOnUiThread
                llExportProgress.visibility = View.GONE
                val btnExportProject = this.activity.findViewById<MaterialButton>(R.id.btnExportProject) ?: return@runOnUiThread
                btnExportProject.setIconResource(R.drawable.export)
            }
            this.activity._active_notification = null
        }

        override fun on_cancel() {
            this.cancelled = true
            this.activity.feedback_msg(this.activity.getString(R.string.export_cancelled))
            this.activity.runOnUiThread {
                val llExportProgress = this.activity.findViewById<View>(R.id.llExportProgress) ?: return@runOnUiThread
                llExportProgress.visibility = View.GONE
                val btnExportProject = this.activity.findViewById<MaterialButton>(R.id.btnExportProject) ?: return@runOnUiThread
                btnExportProject.setIconResource(R.drawable.export)
            }

            val builder = this.activity.get_notification() ?: return
            builder.setContentText(this.activity.getString(R.string.export_cancelled))
                .setProgress(0, 0, false)
                .setAutoCancel(true)
                .setTimeoutAfter(5000)
                .clearActions()

            @SuppressLint("MissingPermission")
            if (this.activity.has_notification_permission()) {
                val notification_manager = NotificationManagerCompat.from(this.activity)
                notification_manager.notify(this.activity.NOTIFICATION_ID, builder.build())
            }
            this.activity._active_notification = null
        }

        override fun on_progress_update(progress: Double) {
            val progress_rounded = ((progress + this.working_y) * 100.0 / this.total_count.toDouble()).roundToInt()
            this.activity.runOnUiThread {
                val tvExportProgress = this.activity.findViewById<TextView>(R.id.tvExportProgress) ?: return@runOnUiThread
                tvExportProgress.text = this.activity.getString(R.string.label_export_progress, progress_rounded)
            }

            val builder = this.activity.get_notification() ?: return
            builder.setProgress(100, progress_rounded, false)

            @SuppressLint("MissingPermission")
            if (this.activity.has_notification_permission()) {
                this.notification_manager.notify(
                    this.activity.NOTIFICATION_ID,
                    builder.build()
                )
            }
        }
        fun update(y: Int, file_uri: Uri) {
            this.working_y = y
            this.file_uri = file_uri
        }
    }

    private var _export_multi_line_wav_intent_launcher = registerForActivityResult(ActivityResultContracts.StartActivityForResult()) { result ->
        if (this._soundfont == null) {
            // Throw Error. Currently unreachable by ui
            return@registerForActivityResult
        }

        this.getNotificationPermission()
        thread {
            if (result.resultCode == RESULT_OK) {
                result?.data?.data?.also { tree_uri ->
                    if (this.view_model.export_handle != null) {
                        return@thread
                    }
                    val directory = DocumentFile.fromTreeUri(this, tree_uri) ?: return@thread
                    val opus_manager_copy = OpusLayerBase()
                    opus_manager_copy.project_change_json(this.get_opus_manager().to_json())

                    var line_count = 0
                    val skip_lines = mutableSetOf<Pair<Int, Int>>()

                    opus_manager_copy.get_all_channels().forEachIndexed channel_loop@{ i: Int, channel: OpusChannelAbstract<*, *> ->
                        channel.lines.forEachIndexed line_loop@{ j: Int, line: OpusLineAbstract<*> ->
                            if (line.muted || channel.muted) {
                                skip_lines.add(Pair(i, j))
                                return@line_loop
                            }

                            var skip = true
                            for (beat in line.beats) {
                                if (!beat.is_eventless())  {
                                    skip = false
                                    break
                                }
                            }

                            if (skip) {
                                skip_lines.add(Pair(i, j))
                            } else {
                                line_count += 1
                            }
                        }
                    }

                    val export_event_handler = MultiExporterEventHandler(this, line_count)

                    var y = 0
                    var c = 0
                    outer@ for (channel in opus_manager_copy.get_all_channels()) {
                        for (l in channel.lines.indices) {
                            if (skip_lines.contains(Pair(c, l))) {
                                continue
                            }

                            val file = directory.createFile("audio/wav", getString(R.string.export_wav_lines_filename, c, l)) ?: continue
                            val file_uri = file.uri

                            /* TMP file is necessary since we can't easily predict the exact frame count. */
                            val tmp_file = File("${this.filesDir}/.tmp_wav_data")
                            if (tmp_file.exists()) {
                                tmp_file.delete()
                            }

                            tmp_file.deleteOnExit()
                            val exporter_sample_handle_manager = SampleHandleManager(this._soundfont!!, 44100, 22050)

                            var c_b = 0
                            for (channel_copy in opus_manager_copy.get_all_channels()) {
                                var l_b = 0
                                for (line_copy in channel_copy.lines) {
                                    if (c_b == c && l_b == l) {
                                        line_copy.unmute()
                                    } else {
                                        line_copy.mute()
                                    }
                                    l_b += 1
                                }
                                c_b += 1
                            }


                            val parcel_file_descriptor = applicationContext.contentResolver.openFileDescriptor(file_uri, "w") ?: continue@outer
                            val output_stream = FileOutputStream(parcel_file_descriptor.fileDescriptor)
                            val buffered_output_stream = BufferedOutputStream(output_stream)
                            val data_output_buffer = DataOutputStream(buffered_output_stream)

                            export_event_handler.update(y++, file_uri)
                            this.view_model.export_wav(
                                opus_manager_copy,
                                exporter_sample_handle_manager,
                                data_output_buffer,
                                tmp_file,
                                this.configuration,
                                export_event_handler,
                                true,
                                true,
                                false
                            )

                            data_output_buffer.close()
                            buffered_output_stream.close()
                            output_stream.close()
                            parcel_file_descriptor.close()
                            tmp_file.delete()

                            if (export_event_handler.cancelled) {
                                break@outer
                            }
                        }
                        c++
                    }
                }
            }
        }
    }


    private var _export_multi_channel_wav_intent_launcher = registerForActivityResult(ActivityResultContracts.StartActivityForResult()) { result ->
        if (this._soundfont == null) {
            // Throw Error. Currently unreachable by ui
            return@registerForActivityResult
        }

        this.getNotificationPermission()
        thread {
            if (result.resultCode == RESULT_OK) {
                result?.data?.data?.also { tree_uri ->
                    if (this.view_model.export_handle != null) {
                        return@thread
                    }
                    val directory = DocumentFile.fromTreeUri(this, tree_uri) ?: return@thread
                    val opus_manager_copy = OpusLayerBase()
                    opus_manager_copy.project_change_json(this.get_opus_manager().to_json())

                    var channel_count = 0
                    val skip_channels = mutableSetOf<Int>()

                    opus_manager_copy.get_all_channels().forEachIndexed channel_loop@{ i: Int, channel: OpusChannelAbstract<*, *> ->
                        if (channel.muted) {
                            skip_channels.add(i)
                            return@channel_loop
                        }

                        var skip = true
                        channel.lines.forEachIndexed line_loop@{ j: Int, line: OpusLineAbstract<*> ->
                            if (line.muted || !skip) {
                                return@line_loop
                            }

                            for (beat in line.beats) {
                                if (!beat.is_eventless())  {
                                    skip = false
                                    return@line_loop
                                }
                            }
                        }

                        if (skip) {
                            skip_channels.add(i)
                        } else {
                            channel_count += 1
                        }
                    }

                    val export_event_handler = MultiExporterEventHandler(this, channel_count)

                    var y = 0
                    var c = 0
                    outer@ for (channel in opus_manager_copy.get_all_channels()) {
                        if (skip_channels.contains(c)) {
                            continue
                        }

                        val file = directory.createFile("audio/wav", getString(R.string.export_wav_channels_filename, c)) ?: continue
                        val file_uri = file.uri

                        /* TMP file is necessary since we can't easily predict the exact frame count. */
                        val tmp_file = File("${this.filesDir}/.tmp_wav_data")
                        if (tmp_file.exists()) {
                            tmp_file.delete()
                        }

                        tmp_file.deleteOnExit()
                        val exporter_sample_handle_manager = SampleHandleManager(this._soundfont!!, 44100, 22050)

                        var c_b = 0
                        for (channel_copy in opus_manager_copy.get_all_channels()) {
                            if (c_b == c) {
                                channel_copy.unmute()
                            } else {
                                channel_copy.mute()
                            }
                            c_b += 1
                        }


                        val parcel_file_descriptor = applicationContext.contentResolver.openFileDescriptor(file_uri, "w") ?: continue@outer
                        val output_stream = FileOutputStream(parcel_file_descriptor.fileDescriptor)
                        val buffered_output_stream = BufferedOutputStream(output_stream)
                        val data_output_buffer = DataOutputStream(buffered_output_stream)

                        export_event_handler.update(y++, file_uri)
                        this.view_model.export_wav(
                            opus_manager_copy,
                            exporter_sample_handle_manager,
                            data_output_buffer,
                            tmp_file,
                            this.configuration,
                            export_event_handler,
                            true,
                            false,
                            false
                        )

                        data_output_buffer.close()
                        buffered_output_stream.close()
                        output_stream.close()
                        parcel_file_descriptor.close()
                        tmp_file.delete()

                        if (export_event_handler.cancelled) {
                            break@outer
                        }
                        c++
                    }
                }
            }
        }
    }

    private var _export_wav_intent_launcher = registerForActivityResult(ActivityResultContracts.StartActivityForResult()) { result ->
        if (this._soundfont == null) {
            // Throw Error. Currently unreachable by ui
            return@registerForActivityResult
        }

        this.getNotificationPermission()
        thread {
            if (result.resultCode == Activity.RESULT_OK) {
                result?.data?.data?.also { uri ->
                    if (this.view_model.export_handle != null) {
                        return@thread
                    }

                    /* TMP file is necessary since we can't easily predict the exact frame count. */
                    val tmp_file = File("${this.filesDir}/.tmp_wav_data")
                    if (tmp_file.exists()) {
                        tmp_file.delete()
                    }

                    tmp_file.deleteOnExit()
                    val exporter_sample_handle_manager = SampleHandleManager(
                        this._soundfont!!,
                        44100,
                        22050
                    )

                    val parcel_file_descriptor = applicationContext.contentResolver.openFileDescriptor(uri, "w") ?: return@thread
                    val output_stream = FileOutputStream(parcel_file_descriptor.fileDescriptor)
                    val buffered_output_stream = BufferedOutputStream(output_stream)
                    val data_output_buffer = DataOutputStream(buffered_output_stream)

                    this.view_model.export_wav(this.get_opus_manager(), exporter_sample_handle_manager, data_output_buffer, tmp_file, this.configuration, object : WavConverter.ExporterEventHandler {
                        val notification_manager = NotificationManagerCompat.from(this@MainActivity)
                        fun close_buffers() {
                            data_output_buffer.close()
                            buffered_output_stream.close()
                            output_stream.close()
                            parcel_file_descriptor.close()
                            tmp_file.delete()
                        }

                        override fun on_start() {
                            this@MainActivity.runOnUiThread {
                                val btnExportProject = this@MainActivity.findViewById<MaterialButton>(R.id.btnExportProject) ?: return@runOnUiThread
                                btnExportProject.setIconResource(R.drawable.baseline_cancel_42)
                                val llExportProgress = this@MainActivity.findViewById<View>(R.id.llExportProgress) ?: return@runOnUiThread
                                llExportProgress.visibility = View.VISIBLE

                                val tvExportProgress = this@MainActivity.findViewById<TextView>(R.id.tvExportProgress) ?: return@runOnUiThread
                                tvExportProgress.text = "0%"
                            }
                            this@MainActivity.feedback_msg(this@MainActivity.getString(R.string.export_wav_feedback))
                            val builder = this@MainActivity.get_notification() ?: return
                            @SuppressLint("MissingPermission")
                            if (this@MainActivity.has_notification_permission()) {
                                this.notification_manager.notify(this@MainActivity.NOTIFICATION_ID, builder.build())
                            }
                        }

                        override fun on_complete() {
                            this.close_buffers()

                            val builder = this@MainActivity.get_notification()
                            if (builder != null) {
                                // NON functional ATM, Open file from notification
                                val go_to_file_intent = Intent()
                                go_to_file_intent.action = Intent.ACTION_VIEW
                                go_to_file_intent.setDataAndType(uri, "*/*")

                                val pending_go_to_intent = PendingIntent.getActivity(
                                    this@MainActivity,
                                    0,
                                    go_to_file_intent,
                                    PendingIntent.FLAG_IMMUTABLE
                                )

                                builder.setContentText(this@MainActivity.getString(R.string.export_wav_notification_complete))
                                    .clearActions()
                                    .setAutoCancel(true)
                                    .setProgress(0, 0, false)
                                    .setTimeoutAfter(5000)
                                    .setSilent(false)
                                    .setContentIntent(pending_go_to_intent)

                                @SuppressLint("MissingPermission")
                                if (this@MainActivity.has_notification_permission()) {
                                    this.notification_manager.notify(this@MainActivity.NOTIFICATION_ID, builder.build())
                                }
                            }

                            this@MainActivity.feedback_msg(this@MainActivity.getString(R.string.export_wav_feedback_complete))

                            this@MainActivity.runOnUiThread {
                                val llExportProgress =
                                    this@MainActivity.findViewById<View>(R.id.llExportProgress)
                                        ?: return@runOnUiThread
                                llExportProgress.visibility = View.GONE
                                val btnExportProject = this@MainActivity.findViewById<MaterialButton>(R.id.btnExportProject) ?: return@runOnUiThread
                                btnExportProject.setIconResource(R.drawable.export)
                            }
                            this@MainActivity._active_notification = null
                        }

                        override fun on_cancel() {
                            this.close_buffers()

                            this@MainActivity.feedback_msg(this@MainActivity.getString(R.string.export_cancelled))
                            this@MainActivity.runOnUiThread {
                                val llExportProgress = this@MainActivity.findViewById<View>(R.id.llExportProgress)
                                        ?: return@runOnUiThread
                                llExportProgress.visibility = View.GONE
                                val btnExportProject = this@MainActivity.findViewById<MaterialButton>(R.id.btnExportProject) ?: return@runOnUiThread
                                btnExportProject.setIconResource(R.drawable.export)
                            }

                            val builder = this@MainActivity.get_notification() ?: return
                            builder.setContentText(this@MainActivity.getString(R.string.export_cancelled))
                                .setProgress(0, 0, false)
                                .setAutoCancel(true)
                                .setTimeoutAfter(5000)
                                .clearActions()

                            @SuppressLint("MissingPermission")
                            if (this@MainActivity.has_notification_permission()) {
                                val notification_manager = NotificationManagerCompat.from(this@MainActivity)
                                notification_manager.notify(this@MainActivity.NOTIFICATION_ID, builder.build())
                            }
                            this@MainActivity._active_notification = null
                        }

                        override fun on_progress_update(progress: Double) {
                            val progress_rounded = (progress * 100.0).roundToInt()
                            this@MainActivity.runOnUiThread {
                                val tvExportProgress = this@MainActivity.findViewById<TextView>(R.id.tvExportProgress) ?: return@runOnUiThread
                                tvExportProgress.text = getString(R.string.label_export_progress, progress_rounded)
                            }

                            val builder = this@MainActivity.get_notification() ?: return
                            builder.setProgress(100, progress_rounded, false)

                            @SuppressLint("MissingPermission")
                            if (this@MainActivity.has_notification_permission()) {
                                this.notification_manager.notify(
                                    this@MainActivity.NOTIFICATION_ID,
                                    builder.build()
                                )
                            }
                        }
                    })

                    exporter_sample_handle_manager.destroy()
                }
            }
        }
    }

    private var _crash_report_intent_launcher = registerForActivityResult(ActivityResultContracts.StartActivityForResult()) { result ->
        val path = this.getExternalFilesDir(null).toString()
        val file = File("$path/bkp_crashreport.log")
        if (result.resultCode == Activity.RESULT_OK) {
            result?.data?.data?.also { uri ->
                val content = file.readText()
                applicationContext.contentResolver.openFileDescriptor(uri, "w")?.use {
                    FileOutputStream(it.fileDescriptor).write(content.toByteArray())
                    file.delete()
                }
            }
        } else {
            file.delete()
        }
    }

    private var _export_project_intent_launcher = registerForActivityResult(ActivityResultContracts.StartActivityForResult()) { result ->
        if (result.resultCode == Activity.RESULT_OK) {
            val opus_manager = this.get_opus_manager()
            result?.data?.data?.also { uri ->
                applicationContext.contentResolver.openFileDescriptor(uri, "w")?.use {
                    val json_string = opus_manager.to_json().to_string()
                    FileOutputStream(it.fileDescriptor).write(json_string.toByteArray())
                    this.feedback_msg(getString(R.string.feedback_exported))
                }
            }
        }
    }

    internal var general_import_intent_launcher = registerForActivityResult(ActivityResultContracts.StartActivityForResult()) { result ->
        if (result.resultCode == Activity.RESULT_OK) {
            result?.data?.data?.also { uri ->
                this.navigate_import(uri)
            }
        }
    }

    private var _export_midi_intent_launcher = registerForActivityResult(ActivityResultContracts.StartActivityForResult()) { result ->
        if (result.resultCode == Activity.RESULT_OK) {
            val opus_manager = this.get_opus_manager()
            result?.data?.data?.also { uri ->
                applicationContext.contentResolver.openFileDescriptor(uri, "w")?.use {
                    FileOutputStream(it.fileDescriptor).write(opus_manager.get_midi().as_bytes())
                    this.feedback_msg(getString(R.string.feedback_exported_to_midi))
                }
            }
        }
    }

    var _import_soundfont_intent_listener = registerForActivityResult(ActivityResultContracts.StartActivityForResult()) { result ->
        if (result.resultCode == Activity.RESULT_OK) {
            result?.data?.data?.also { uri ->
                if (uri.path != null) {
                    val soundfont_dir = this.get_soundfont_directory()
                    val file_name = this.parse_file_name(uri)

                    val new_file = File("${soundfont_dir}/$file_name")
                    this.applicationContext.contentResolver.openFileDescriptor(uri, "r")?.use {
                        try {
                            new_file.outputStream().use { output_stream: FileOutputStream ->
                                FileInputStream(it.fileDescriptor).use { input_stream: FileInputStream ->
                                    input_stream.copyTo(output_stream, 4096 * 4)
                                }
                            }
                        } catch (e: FileNotFoundException) {
                            // TODO:  Feedback? Only breaks on devices without properly implementation (realme RE549c)
                        }
                    }

                    try {
                        SoundFont(new_file.path)
                        this.get_action_interface().ignore().set_soundfont(new_file.name)
                    } catch (e: Exception) {
                        this.feedback_msg(getString(R.string.feedback_invalid_sf2_file))
                        new_file.delete()
                        return@registerForActivityResult
                    }

                    // Hide the warning
                    if (this.get_active_fragment() is FragmentGlobalSettings && this.is_soundfont_available()) {
                        this.findViewById<LinearLayout>(R.id.llSFWarning).visibility = View.GONE
                    }
                } else {
                    throw FileNotFoundException()
                }
            }
        }
    }

    override fun onConfigurationChanged(newConfig: Configuration) {
        super.onConfigurationChanged(newConfig)
        this.recreate()
    }

    override fun onPause() {
        this.playback_stop()
        this.unregisterReceiver(this.broadcast_receiver)
        this.playback_stop_midi_output()
        this._midi_interface.close_connected_devices()
        this._binding.appBarMain.toolbar.hideOverflowMenu()
        super.onPause()
    }

    override fun onKeyDown(key_code: Int, event: KeyEvent?): Boolean {
        val active_fragment = this.get_active_fragment()
        val cancel_super = if (event != null) {
            when (active_fragment) {
                is FragmentEditor ->
                    try {
                        active_fragment.keyboard_input_interface?.input(key_code, event) ?: false
                    } catch (e: Exception) {
                        true
                    }
                else -> false
            }
        } else {
            false
        }

        return if (cancel_super) {
            true
        } else {
            super.onKeyDown(key_code, event)
        }
    }

    override fun onResume() {
        super.onResume()
        this.drawer_lock()

        registerReceiver(
            this.broadcast_receiver,
            this.receiver_intent_filter,
            if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.TIRAMISU) {
                RECEIVER_NOT_EXPORTED
            } else {
                0
            }
        )

        if (this._midi_playback_device != null) {
            this.playback_state_soundfont = PlaybackState.Ready
        }

        if (this.is_connected_to_physical_device()) {
            this.playback_state_midi = PlaybackState.Ready
        }
    }

    fun save_to_backup() {
        val opus_manager = this.get_opus_manager()
        if (!opus_manager.first_load_done) {
            return
        }
        val path = opus_manager.path
        if (path != null) {
            val path_file = File("${applicationInfo.dataDir}/.bkp_path")
            path_file.writeText(path)
        }
        opus_manager.save("${applicationInfo.dataDir}/.bkp.json")

        // saving changes the path, need to change it back
        opus_manager.path = path
    }

    override fun onSaveInstanceState(outState: Bundle) {
        // Can't reliably put json in outstate. there is a size limit
        this.save_to_backup()
        super.onSaveInstanceState(outState)
    }

    override fun onCreate(savedInstanceState: Bundle?) {
       super.onCreate(savedInstanceState)
        this.check_for_crash_report()

        Thread.setDefaultUncaughtExceptionHandler { paramThread, paramThrowable ->
            Log.d("pagandebug", "$paramThrowable")
            if (this@MainActivity.is_debug_on()) {
                this@MainActivity.save_actions()
            }
            this@MainActivity.save_to_backup()
            this@MainActivity.bkp_crash_report(paramThrowable)

            val ctx = applicationContext
            val pm = ctx.packageManager
            val intent = pm.getLaunchIntentForPackage(ctx.packageName)
            val mainIntent = Intent.makeRestartActivityTask(intent!!.component)
            ctx.startActivity(mainIntent)
            Runtime.getRuntime().exit(0)

        }

        this._config_path = "${this.getExternalFilesDir(null)}/pagan.cfg"
        // [Re]move config file from < v1.1.2
        val old_config_file = File("${applicationInfo.dataDir}/pagan.cfg")
        val new_config_file = File(this._config_path)
        if (old_config_file.exists()) {
            if (!new_config_file.exists()) {
                old_config_file.copyTo(new_config_file)
            }
            old_config_file.delete()
        }

        this.configuration = try {
            PaganConfiguration.from_path(this._config_path)
        } catch (e: Exception) {
            PaganConfiguration()
        }

        this._midi_interface = object : MidiController(this) {
            override fun onDeviceAdded(device_info: MidiDeviceInfo) {
                if (!this@MainActivity.update_playback_state_midi(PlaybackState.Ready)) {
                    return
                }

                when (this@MainActivity.playback_state_soundfont) {
                    PlaybackState.Playing,
                    PlaybackState.Queued -> {
                        this@MainActivity.playback_stop()
                    }

                    else -> { /* pass */ }
                }

                this@MainActivity.runOnUiThread {
                    this@MainActivity.update_menu_options()
                    if (!this@MainActivity.configuration.allow_midi_playback) {
                        return@runOnUiThread
                    }

                    this@MainActivity.setup_project_config_drawer_export_button()

                    val channel_recycler = this@MainActivity.findViewById<ChannelOptionRecycler>(R.id.rvActiveChannels)
                    // Should always be null since this can only be changed from a different menu
                    if (channel_recycler.adapter != null) {
                        val channel_adapter = channel_recycler.adapter as ChannelOptionAdapter
                        channel_adapter.notify_soundfont_changed()
                    }
                    this@MainActivity.populate_active_percussion_names(true)
                }

                if (this@MainActivity.get_opus_manager().is_tuning_standard()) {
                    this@MainActivity.disconnect_feedback_device()
                }
            }

            override fun onDeviceRemoved(device_info: MidiDeviceInfo) {
                when (this@MainActivity.playback_state_midi) {
                    PlaybackState.Playing,
                    PlaybackState.Queued -> {
                        this@MainActivity.playback_stop_midi_output()
                    }

                    else -> { /* pass */ }
                }

                // Kludge. need a sleep to give output devices a chance to disconnect
                Thread.sleep(1000)

                this@MainActivity.runOnUiThread {
                    this@MainActivity.update_menu_options()
                    if (!this@MainActivity.is_connected_to_physical_device()) {
                        this@MainActivity.setup_project_config_drawer_export_button()

                        val channel_recycler = this@MainActivity.findViewById<ChannelOptionRecycler>(R.id.rvActiveChannels)
                        // Should always be null since this can only be changed from a different menu
                        if (channel_recycler.adapter != null) {
                            val channel_adapter = channel_recycler.adapter as ChannelOptionAdapter
                            channel_adapter.notify_soundfont_changed()
                        }

                        this@MainActivity.populate_active_percussion_names(true)
                    }
                }
            }
        }

        if (!this.configuration.allow_midi_playback) {
            this.block_physical_midi_output()
        }

        this._midi_interface.connect_virtual_input_device(this._virtual_input_device)

        // Listens for SongPositionPointer (provided by midi) and scrolls to that beat
        this._midi_interface.connect_virtual_output_device(object : VirtualMidiOutputDevice {
            override fun onSongPositionPointer(event: SongPositionPointer) {
                if (event.get_beat() >= this@MainActivity.get_opus_manager().length) {
                    return
                }
                this@MainActivity.get_opus_manager().cursor_select_column(event.get_beat())
                // Force scroll here, cursor_select_column doesn't scroll if the column is already visible
                this@MainActivity.runOnUiThread {
                    this@MainActivity.findViewById<EditorTable>(R.id.etEditorTable)
                        ?.scroll_to_position(x = event.get_beat(), force = true)
                }
            }
        })

        this._midi_interface.connect_virtual_input_device(this._midi_feedback_dispatcher)

        this._project_manager = ProjectManager(this)
        // Move files from applicationInfo.data to externalfilesdir (pre v1.1.2 location)
        val old_projects_dir = File("${applicationInfo.dataDir}/projects")
        if (old_projects_dir.isDirectory) {
            for (f in old_projects_dir.listFiles()!!) {
                val new_file_name = this._project_manager.get_new_path()
                f.copyTo(File(new_file_name))
            }
            old_projects_dir.deleteRecursively()
        }

        this.requestedOrientation = this.configuration.force_orientation

        this._binding = ActivityMainBinding.inflate(this.layoutInflater)
        this.setContentView(this._binding.root)
        this.setSupportActionBar(this._binding.appBarMain.toolbar)
        this._binding.root.setBackgroundColor(resources.getColor(R.color.main_bg))

        this.view_model.action_interface.attach_activity(this)
        this.view_model.opus_manager.attach_activity(this)

        val toolbar = this._binding.appBarMain.toolbar

        toolbar.setOnLongClickListener {
            this.get_action_interface().set_project_name_and_notes()
            true
        }

        toolbar.background = null

        //////////////////////////////////////////
        if (this.configuration.soundfont != null) {
            val path = "${this.getExternalFilesDir(null)}/SoundFonts/${this.configuration.soundfont}"
            val sf_file = File(path)
            if (sf_file.exists()) {
                try {
                    this._soundfont = SoundFont(path)
                    this.populate_supported_soundfont_instrument_names()
                    this._sample_handle_manager = SampleHandleManager(
                        this._soundfont!!,
                        this.configuration.sample_rate,
                        this.configuration.sample_rate, // Use Large buffer
                        ignore_lfo = true
                    )

                    this._midi_playback_device = PlaybackDevice(
                        this,
                        this._sample_handle_manager!!,
                        WaveGenerator.StereoMode.Stereo
                    )

                    if (!this.is_connected_to_physical_device()) {
                        val buffer_size = this.configuration.sample_rate / 4
                        this._feedback_sample_manager = SampleHandleManager(
                            this._soundfont!!,
                            this.configuration.sample_rate,
                            buffer_size - 2 + (if (buffer_size % 2 == 0) { 2 } else { 1 })
                            //sample_limit = this.configuration.playback_sample_limit,
                            //ignore_envelopes_and_lfo = true
                        )
                    }
                } catch (e: Riff.InvalidRiff) {
                    this.configuration.soundfont = null
                    // Invalid soundfont somehow set
                }
            }
        }

        this.update_channel_instruments(this.get_opus_manager().channels.size)
        ///////////////////////////////////////////

        onBackPressedDispatcher.addCallback(this, object : OnBackPressedCallback(true) {
            override fun handleOnBackPressed() {
                this@MainActivity.get_action_interface().go_back()
            }
        })

        val drawer_layout = this.findViewById<DrawerLayout>(R.id.drawer_layout) ?: return
        drawer_layout.addDrawerListener(
            object : ActionBarDrawerToggle( this, drawer_layout, R.string.drawer_open, R.string.drawer_close) {
                override fun onDrawerOpened(drawerView: View) {
                    this@MainActivity.get_action_interface().track(ActionTracker.TrackedAction.DrawerOpen)
                    val channel_recycler = this@MainActivity.findViewById<ChannelOptionRecycler>(R.id.rvActiveChannels)
                    if (channel_recycler.adapter == null) {
                        ChannelOptionAdapter(this@MainActivity.get_opus_manager(), channel_recycler)
                    }
                    super.onDrawerOpened(drawerView)

                    val channel_adapter = (channel_recycler.adapter as ChannelOptionAdapter)
                    if (channel_adapter.itemCount == 0) {
                        channel_adapter.setup()
                    }

                    this@MainActivity.playback_stop()
                    this@MainActivity.playback_stop_midi_output()
                    this@MainActivity.drawer_unlock() // So the drawer can be closed with a swipe
                }

                override fun onDrawerClosed(drawerView: View) {
                    this@MainActivity.get_action_interface().track(ActionTracker.TrackedAction.DrawerClose)
                    super.onDrawerClosed(drawerView)
                    this@MainActivity.drawer_lock() // so the drawer can't be opened with a swipe
                }
            }
        )
    }



    override fun onSupportNavigateUp(): Boolean {
        val navController = this.findNavController(R.id.nav_host_fragment_content_main)
        return navController.navigateUp(this._app_bar_configuration) || super.onSupportNavigateUp()
    }

    override fun onCreateOptionsMenu(menu: Menu): Boolean {
        this.menuInflater.inflate(R.menu.main_options_menu, menu)
        this._options_menu = menu
        val output = super.onCreateOptionsMenu(menu)
        this.update_menu_options()
        return output
    }

    override fun onOptionsItemSelected(item: MenuItem): Boolean {
        if (item.itemId != R.id.itmPlay && item.itemId != R.id.itmPlayMidiOutput) {
            this.playback_stop()
            this.playback_stop_midi_output()
        }

        when (item.itemId) {
            android.R.id.home -> {
                val fragment = this.get_active_fragment()
                if (fragment is FragmentEditor) {
                    this.drawer_open()
                } else {
                    this.get_action_interface().go_back()
                }
            }

            R.id.itmNewProject -> {
                this.dialog_save_project {
                    this.get_action_interface().new_project()
                }
            }

            R.id.itmLoadProject -> {
                this.dialog_save_project {
                    this.dialog_load_project()
                }
            }

            R.id.itmImportMidi -> {
                this.dialog_save_project {
                    this.select_import_file()
                }
            }

            R.id.itmUndo -> {
                this.get_action_interface().apply_undo()
            }

            R.id.itmPlay -> {
                when (this.playback_state_soundfont) {
                    PlaybackState.Ready -> {
                        this.playback_start()
                    }
                    PlaybackState.Queued,
                    PlaybackState.Playing -> {
                        this.playback_stop()
                    }

                    else -> { }
                }
            }

            R.id.itmPlayMidiOutput -> {
                when (this.playback_state_midi) {
                    PlaybackState.Ready -> {
                        this.playback_start_midi_output()
                    }
                    PlaybackState.Queued,
                    PlaybackState.Playing -> {
                        this.playback_stop_midi_output()
                    }
                    else -> { /* pass */ }
                }
            }

            R.id.itmSettings -> {
                this.get_action_interface().open_settings()
            }
            R.id.itmAbout -> {
                this.get_action_interface().open_about()
            }
            R.id.itmDebug -> {
                val tracker = this.get_action_interface()
                this.save_actions()
                this.feedback_msg("SAVED ACTIONS")

            }
        }
        return super.onOptionsItemSelected(item)
    }

    fun project_save() {
        this._project_manager.save(this.get_opus_manager())
        this.feedback_msg(getString(R.string.feedback_project_saved))
        this.update_menu_options()
    }

    fun project_delete() {
        val title = this.get_opus_manager().project_name ?: getString(R.string.untitled_opus)
        this._project_manager.delete(this.get_opus_manager())

        val fragment = this.get_active_fragment()
        fragment?.setFragmentResult(IntentFragmentToken.New.name, bundleOf())
        if (fragment !is FragmentEditor) {
            this.navigate(R.id.EditorFragment)
        }

        this.feedback_msg(resources.getString(R.string.feedback_delete, title))
    }

    fun project_move_to_copy() {
        this.dialog_save_project {
            this._project_manager.move_to_copy(this.get_opus_manager())
            this.update_title_text()
            this.feedback_msg(getString(R.string.feedback_on_copy))

            val btnDeleteProject = this.findViewById<View>(R.id.btnDeleteProject)
            val btnCopyProject = this.findViewById<View>(R.id.btnCopyProject)
            btnDeleteProject.isEnabled = false
            btnCopyProject.isEnabled = false
        }
    }

    @SuppressLint("ClickableViewAccessibility")
    private fun _enable_blocker_view() {
        val blocker_view = this.findViewById<LinearLayout>(R.id.llClearOverlay)

        if (blocker_view != null && blocker_view.visibility != View.VISIBLE) {
            blocker_view.setOnTouchListener { _, motion_event ->
                /* Allow Scrolling on the y axis when scrolling in the main_recycler */
                if (motion_event == null) {
                } else if (motion_event.action == 1) {
                    this._blocker_scroll_y = null
                } else if (motion_event.action != MotionEvent.ACTION_MOVE) {
                } else {
                    val fragment = this.get_active_fragment()
                    if (fragment !is FragmentEditor) {
                        return@setOnTouchListener false
                    }
                    val editor_table = this.findViewById<EditorTable>(R.id.etEditorTable)
                    val scroll_view = editor_table.get_scroll_view()

                    if (this._blocker_scroll_y == null) {
                        this._blocker_scroll_y = (motion_event.y - scroll_view.y)
                    }

                    val rel_y = (motion_event.y - scroll_view.y)
                    val delta_y = this._blocker_scroll_y!! - rel_y

                    scroll_view.scrollBy(0, delta_y.toInt())
                    this._blocker_scroll_y = rel_y
                }
                true
            }

            blocker_view.visibility = View.VISIBLE
            this.window.addFlags(WindowManager.LayoutParams.FLAG_KEEP_SCREEN_ON)
        }
    }

    private fun _disable_blocker_view() {
        // Assume playback states have been updated BEFORE calling this function
        // Need to check both since they both use the same blocker, though it should never be an issue
        if (this.playback_state_midi == PlaybackState.Playing || this.playback_state_soundfont == PlaybackState.Playing) {
            return
        }
        this.window.clearFlags(WindowManager.LayoutParams.FLAG_KEEP_SCREEN_ON)
        val blocker_view = this.findViewById<LinearLayout>(R.id.llClearOverlay) ?: return
        blocker_view.visibility = View.GONE
    }

    private fun playback_start() {
        if (!this.update_playback_state_soundfont(PlaybackState.Queued)) {
            this.feedback_msg(getString(R.string.playback_failed))
            return
        }

        this._enable_blocker_view()
        this.runOnUiThread {
            this.set_playback_button(R.drawable.baseline_play_disabled_24)
            this.loading_reticle_show(getString(R.string.reticle_msg_start_playback))
        }

        var start_point = this.get_working_column()
        if (start_point >= this.get_opus_manager().length - 1) {
            start_point = 0
        }
        // Currently, Midi2.0 output is not supported. will be needed for N-radix projects
        thread {
            // May Not need to set instruments at beginning of playback, commenting for now
            //val opus_manager = this.get_opus_manager()
            //opus_manager.channels.forEachIndexed { i: Int, channel: OpusChannel ->
            //    val (bank, program) = channel.get_instrument()
            //    opus_manager.sample_handle_manager?.select_bank(channel.midi_channel, bank)
            //    opus_manager.sample_handle_manager?.change_program(channel.midi_channel, program)
            //}

            this._midi_playback_device?.play_opus(start_point)
        }
    }

    private fun playback_start_midi_output() {
        if (!this.update_playback_state_midi(PlaybackState.Queued)) {
            this.feedback_msg(getString(R.string.playback_failed))
            return
        }

        this.loading_reticle_show(getString(R.string.reticle_msg_start_playback))
        this._enable_blocker_view()

        var start_point = this.get_working_column()
        val opus_manager = this.get_opus_manager()
        if (start_point >= opus_manager.length - 1) {
            start_point = 0
        }

        val midi = opus_manager.get_midi(start_point)

        this.runOnUiThread {
            this.loading_reticle_hide()
            this.set_midi_playback_button(R.drawable.ic_baseline_pause_24)
        }

        if (!this.update_playback_state_midi(PlaybackState.Playing)) {
            this.restore_midi_playback_state()
            return
        }

        thread {
            try {
                this._midi_interface.open_connected_devices()
                this._virtual_input_device.play_midi(midi) {
                    this.runOnUiThread {
                        this.playback_stop_midi_output()
                    }
                }
            } catch (e: java.io.IOException) {
                this.runOnUiThread {
                    this.playback_stop_midi_output()
                }
            }
        }
    }

    internal fun playback_stop() {
        if (this.update_playback_state_soundfont(PlaybackState.Stopping)) {
            this.loading_reticle_hide()
            this._midi_playback_device?.kill()
        }
    }

    internal fun playback_stop_midi_output() {
        if (this.update_playback_state_midi(PlaybackState.Stopping)) {
            this.loading_reticle_hide()
            this._virtual_input_device.stop()
            this.restore_midi_playback_state()
        }
    }

    fun restore_playback_state() {
        if (this.update_playback_state_soundfont(PlaybackState.Ready)) {
            this.runOnUiThread {
                this.set_playback_button(R.drawable.ic_baseline_play_arrow_24)
                this._disable_blocker_view()
            }
        }
    }

    fun restore_midi_playback_state() {
        if (this.update_playback_state_midi(PlaybackState.Ready)) {
            this.runOnUiThread {
                this.set_midi_playback_button(R.drawable.ic_baseline_play_arrow_24)
                this._disable_blocker_view()
            }
        }
    }

    fun get_new_project_path(): String {
        return this._project_manager.get_new_path()
    }

    // Ui Wrappers ////////////////////////////////////////////
    private fun drawer_close() {
        this.get_action_interface().drawer_close()
    }

    private fun drawer_open() {
        this.get_action_interface().drawer_open()
    }

    fun drawer_lock() {
        this._binding.root.setDrawerLockMode(DrawerLayout.LOCK_MODE_LOCKED_CLOSED)
    }

    fun drawer_unlock() {
        try {
            this._binding.root.setDrawerLockMode(DrawerLayout.LOCK_MODE_UNLOCKED)
            this.findViewById<LinearLayout>(R.id.config_drawer)?.refreshDrawableState()
        } catch (e: UninitializedPropertyAccessException) {
            // pass, if it's not initialized, it's not locked
        }
    }

    fun feedback_msg(msg: String) {
        this.runOnUiThread {
            Toast.makeText(this, msg, Toast.LENGTH_SHORT).show()
        }
    }

    fun loading_reticle_show(title_msg: String? = null) {
        this.runOnUiThread {
            if (title_msg != null) {
                this.force_title_text(title_msg)
            }

            if (this._progress_bar == null) {
                this._progress_bar = LayoutInflater.from(this)
                    .inflate(
                        R.layout.loading_reticle,
                        this._binding.root as ViewGroup,
                        false
                    ) as ConstraintLayout
            }

            this._progress_bar!!.isClickable = true
            val parent = this._progress_bar!!.parent
            if (parent != null) {
                (parent as ViewGroup).removeView(this._progress_bar)
            }

            try {
                this._binding.root.addView(this._progress_bar)
            } catch (e: UninitializedPropertyAccessException) {
                // pass
            }
        }
    }

    fun loading_reticle_hide() {
        thread {
            this.runOnUiThread {
                this.clear_forced_title()
                val progressBar = this._progress_bar ?: return@runOnUiThread
                if (progressBar.parent != null) {
                    (progressBar.parent as ViewGroup).removeView(progressBar)
                }
            }
        }
    }

    fun navigate(fragment: Int) {
        val navController = findNavController(R.id.nav_host_fragment_content_main)
        if (fragment == R.id.EditorFragment) {
            this._has_seen_front_page = true
        }

        navController.navigate(fragment)
    }


    fun get_active_fragment(): Fragment? {
        val navHost = supportFragmentManager.findFragmentById(R.id.nav_host_fragment_content_main)
        return navHost?.childFragmentManager?.fragments?.get(0)
    }

    fun update_title_text() {
        this.set_title_text(
            if (this._forced_title_text != null) {
                this._forced_title_text!!
            } else {
                when (this.get_active_fragment()) {
                    is FragmentGlobalSettings -> {
                        resources.getString(R.string.settings_fragment_label)
                    }
                    is FragmentLicense,
                    is FragmentLandingPage -> {
                        "${getString(R.string.app_name)} ${getString(R.string.app_version)}"
                    }
                    else -> {
                        this.get_opus_manager().project_name ?: getString(R.string.untitled_opus)
                    }
                }
            }
        )
        this._refresh_toolbar()
    }

    fun set_title_text(new_text: String) {
        this._binding.appBarMain.toolbar.title = new_text
    }

    fun force_title_text(msg: String) {
        this._forced_title_text = msg
        this.update_title_text()
    }

    fun clear_forced_title() {
        this._forced_title_text = null
        this.update_title_text()
    }

    fun update_menu_options() {
        val navHost = supportFragmentManager.findFragmentById(R.id.nav_host_fragment_content_main)
        val options_menu = this._options_menu ?: return

        when (navHost?.childFragmentManager?.fragments?.get(0)) {
            is FragmentEditor -> {
                val play_midi_visible = this.configuration.allow_midi_playback && (this._midi_interface.output_devices_connected() && this.get_opus_manager().is_tuning_standard())
                options_menu.findItem(R.id.itmLoadProject).isVisible = this.has_projects_saved()
                options_menu.findItem(R.id.itmUndo).isVisible = true
                options_menu.findItem(R.id.itmNewProject).isVisible = true
                options_menu.findItem(R.id.itmPlay).isVisible = this._soundfont != null && ! play_midi_visible
                options_menu.findItem(R.id.itmPlayMidiOutput).isVisible = play_midi_visible
                options_menu.findItem(R.id.itmImportMidi).isVisible = true
                options_menu.findItem(R.id.itmSettings).isVisible = true
                options_menu.findItem(R.id.itmAbout).isVisible = true
                options_menu.findItem(R.id.itmDebug).isVisible = this.is_debug_on()
            }
            else -> {
                options_menu.findItem(R.id.itmLoadProject).isVisible = false
                options_menu.findItem(R.id.itmUndo).isVisible = false
                options_menu.findItem(R.id.itmNewProject).isVisible = false
                options_menu.findItem(R.id.itmPlay).isVisible = false
                options_menu.findItem(R.id.itmPlayMidiOutput).isVisible = false
                options_menu.findItem(R.id.itmImportMidi).isVisible = false
                options_menu.findItem(R.id.itmSettings).isVisible = false
                options_menu.findItem(R.id.itmAbout).isVisible = false
                options_menu.findItem(R.id.itmDebug).isVisible = false
            }
        }
    }

    fun setup_project_config_drawer() {
        val opus_manager = this.get_opus_manager()
        val tvChangeProjectName: MaterialButton = this.findViewById(R.id.btnChangeProjectName)
        tvChangeProjectName.setOnClickListener {
            this.get_action_interface().set_project_name_and_notes()
        }

        //-------------------------------------------
        val btnRadix: MaterialButton = this.findViewById(R.id.btnRadix)
        btnRadix.setOnClickListener {
            this.get_action_interface().set_tuning_table_and_transpose()
        }
        //-------------------------------------------
        this.findViewById<View>(R.id.btnAddChannel).setOnClickListener {
            this.get_action_interface().insert_channel(
                opus_manager.channels.size
            )
        }

        this.setup_project_config_drawer_export_button()
        this.findViewById<View>(R.id.btnSaveProject).setOnClickListener {
            if (!it.isEnabled) {
                return@setOnClickListener
            }
            this.get_action_interface().save()
        }
        this.findViewById<View>(R.id.btnSaveProject).setOnLongClickListener {
            if (!it.isEnabled) {
                return@setOnLongClickListener false
            }

            this.export_project()
            true
        }

        val btnDeleteProject = this.findViewById<View>(R.id.btnDeleteProject)
        val btnCopyProject = this.findViewById<View>(R.id.btnCopyProject)
        val file_exists = (opus_manager.path != null && File(opus_manager.path!!).isFile)
        btnDeleteProject.isEnabled = file_exists
        btnCopyProject.isEnabled = file_exists

        btnDeleteProject.setOnClickListener {
            if (it.isEnabled) {
                this.dialog_delete_project()
            }
        }

        btnCopyProject.setOnClickListener {
            if (it.isEnabled) {
                this.get_action_interface().project_copy()
            }
        }
    }

    internal fun _build_dialog_title_view(text: String): TextView {
        val output = TextView(ContextThemeWrapper(this, R.style.dialog_title))
        output.text = text
        return output
    }

    internal fun setup_project_config_drawer_export_button() {
        val export_options = this.get_exportable_options()
        val export_button = this.findViewById<MaterialButton>(R.id.btnExportProject) ?: return
        val export_progress_wrapper = this.findViewById<LinearLayout>(R.id.llExportProgress) ?: return
        if (!this.view_model.is_exporting()) {
            export_button.setIconResource(R.drawable.export)
            export_progress_wrapper.visibility = View.GONE
        } else {
            export_button.setIconResource(R.drawable.baseline_cancel_42)
            export_progress_wrapper.visibility = View.VISIBLE
        }

        if (export_options.isNotEmpty()) {
            export_button.setOnClickListener {
                if (!this.view_model.is_exporting()) {
                    this.dialog_popup_menu(
                        getString(R.string.dlg_export),
                        export_options,
                        default = null
                    ) { _: Int, value: Int ->
                        when (value) {
                            0 -> this.export_project()
                            1 -> this.export_midi()
                            2 -> this.export_wav() // DEBUG
                            3 -> this.export_multi_lines_wav() // DEBUG
                            4 -> this.export_multi_channels_wav() // DEBUG
                        }
                    }
                } else {
                    this.export_wav_cancel()
                }
            }
            export_button.visibility = View.VISIBLE
        } else {
            export_button.visibility = View.GONE
        }
    }

    private fun get_exportable_options(): List<Pair<Int, String>> {
        val export_options = mutableListOf<Pair<Int, String>>(
            Pair(0, getString(R.string.export_option_json))
        )
        if (this.get_opus_manager().is_tuning_standard()) {
            export_options.add( Pair(1, getString(R.string.export_option_midi)) )
        }

        if (this.get_soundfont() != null) {
            export_options.add( Pair(2, getString(R.string.export_option_wav)) )
            export_options.add( Pair(3, getString(R.string.export_option_wav_lines)) )
            export_options.add( Pair(4, getString(R.string.export_option_wav_channels)) )
        }

        return export_options
    }

    // Ui Wrappers End ////////////////////////////////////////

    private fun _get_default_drum_options(): List<Pair<String, Int>> {
        val midi_drums = this.resources.getStringArray(R.array.midi_drums)
        return List(midi_drums.size) { i: Int ->
            Pair(midi_drums[i]!!, i + 27)
        }
    }

    private fun get_drum_options(): List<Pair<String, Int>> {
        if (this._sample_handle_manager == null || this.is_connected_to_physical_device()) {
            return this._get_default_drum_options()
        }

        val preset = try {
            this._sample_handle_manager!!.get_preset(9) ?: return this._get_default_drum_options()
        } catch (e: SoundFont.InvalidPresetIndex) {
            return this._get_default_drum_options()
        }

        val available_drum_keys = mutableSetOf<Pair<String, Int>>()
        for ((_, preset_instrument) in preset.instruments) {
            if (preset_instrument.instrument == null) {
                continue
            }

            val instrument_range = preset_instrument.key_range ?: Pair(0, 127)

            for (sample_directive in preset_instrument.instrument!!.sample_directives.values) {
                val key_range = sample_directive.key_range ?: Pair(0, 127)
                val usable_range = max(key_range.first, instrument_range.first).. min(key_range.second, instrument_range.second)

                var name = sample_directive.sample!!.first().name
                if (name.contains("(")) {
                    name = name.substring(0, name.indexOf("("))
                }

                for (key in usable_range) {
                    var use_name = if (usable_range.first != usable_range.last) {
                        "$name - ${(key - usable_range.first) + 1}"
                    } else {
                        name
                    }
                    available_drum_keys.add(Pair(use_name, key))
                }
            }
        }

        return available_drum_keys.sortedBy {
            it.second
        }
    }

    fun update_channel_instrument(midi_channel: Int, instrument: Pair<Int, Int>) {
        val (midi_bank, midi_program) = instrument
        this._midi_interface.broadcast_event(BankSelect(midi_channel, midi_bank))
        this._midi_interface.broadcast_event(ProgramChange(midi_channel, midi_program))
        if (this._feedback_sample_manager != null) {
            this._feedback_sample_manager!!.select_bank(
                midi_channel,
                midi_bank,
            )
            this._feedback_sample_manager!!.change_program(
                midi_channel,
                midi_program,
            )
        }

        // Don't need to update anything but percussion in the sample_handle_manager
        if (this._sample_handle_manager != null) {
            this._sample_handle_manager!!.select_bank(
                midi_channel,
                midi_bank
            )
            this._sample_handle_manager!!.change_program(
                midi_channel,
                midi_program
            )
        }
    }

    // Update peripheral device instruments, ie feedback device and midi devices
    fun update_channel_instruments(index: Int? = null) {
        val opus_manager = this.get_opus_manager()
        if (index == null) {
            this._feedback_sample_manager?.let { handle_manager: SampleHandleManager ->
                for (channel in opus_manager.get_all_channels()) {
                    val midi_channel = channel.get_midi_channel()
                    val (midi_bank, midi_program) = channel.get_instrument()
                    this._midi_interface.broadcast_event(BankSelect(midi_channel, midi_bank))
                    this._midi_interface.broadcast_event(ProgramChange(midi_channel, midi_program))

                    handle_manager.select_bank(midi_channel, midi_bank)
                    handle_manager.change_program(midi_channel, midi_program)
                }
            }

            this._sample_handle_manager?.let { handle_manager: SampleHandleManager ->
                // Don't need to update anything but percussion here
                val midi_channel = opus_manager.percussion_channel.get_midi_channel()
                val (midi_bank, midi_program) = opus_manager.percussion_channel.get_instrument()
                handle_manager.select_bank(midi_channel, midi_bank)
                handle_manager.change_program(midi_channel, midi_program)
            }
        } else {
            val opus_channel = opus_manager.get_channel(index)
            this.update_channel_instrument(
                opus_channel.get_midi_channel(),
                opus_channel.get_instrument()
            )
        }
    }

    fun get_opus_manager(): OpusManager {
        return this.view_model.opus_manager
    }

    fun play_event(channel: Int, event_value: Int, velocity: Float = .6F) {
        if (event_value < 0) {
            return // No sound to play
        }
        if (!this.is_connected_to_physical_device()) {
            if (this._feedback_sample_manager == null) {
                this.connect_feedback_device()
                this.update_channel_instruments()
            }
        } else {
            this.disconnect_feedback_device()
            this._midi_interface.open_output_devices()
        }

        val opus_manager = this.get_opus_manager()
        val midi_channel = opus_manager.get_channel(channel).get_midi_channel()

        val radix = opus_manager.tuning_map.size
        val (note, bend) = if (opus_manager.is_percussion(channel)) { // Ignore the event data and use percussion map
            Pair(event_value + 27, 0)
        } else {
            val octave = event_value / radix
            val offset = opus_manager.tuning_map[event_value % radix]

            val transpose_offset = 12.0 * opus_manager.transpose.first.toDouble() / opus_manager.transpose.second.toDouble()
            val std_offset = 12.0 * offset.first.toDouble() / offset.second.toDouble()

            val bend = (((std_offset - floor(std_offset)) + (transpose_offset - floor(transpose_offset))) * 512.0).toInt()
            val new_note = (octave * 12) + std_offset.toInt() + transpose_offset.toInt() + 21

            Pair(new_note, bend)
        }

        if (note > 127) {
            return
        }


        this._feedback_sample_manager?.let { handle_manager : SampleHandleManager ->
            if (this._temporary_feedback_devices[this._current_feedback_device] == null) {
                this._temporary_feedback_devices[this._current_feedback_device] = FeedbackDevice(this._feedback_sample_manager!!)
            }

            val event = NoteOn79(
                index=0,
                channel=midi_channel,
                note=note,
                bend=bend,
                velocity = (velocity * 127F).toInt() shl 8,
            )

            this._temporary_feedback_devices[this._current_feedback_device]!!.new_event(event, 250)
            this._current_feedback_device = (this._current_feedback_device + 1) % this._temporary_feedback_devices.size
        } ?: {
            try {
                this._midi_feedback_dispatcher.play_note(
                    midi_channel,
                    note,
                    bend,
                    (velocity * 127F).toInt(),
                    !opus_manager.is_tuning_standard() || !this.is_connected_to_physical_device()
                )
            } catch (e: VirtualMidiInputDevice.DisconnectedException) {
                // Feedback shouldn't be necessary here. But i'm sure that'll come back to bite me
            }
        }()
    }

    fun import_project(path: String) {
        this.applicationContext.contentResolver.openFileDescriptor(path.toUri(), "r")?.use {
            val bytes = FileInputStream(it.fileDescriptor).readBytes()
            this.get_opus_manager().load(bytes, this._project_manager.get_new_path())
        }
    }

    fun import_midi(path: String) {
        val bytes = this.applicationContext.contentResolver.openFileDescriptor(path.toUri(), "r")?.use {
            FileInputStream(it.fileDescriptor).readBytes()
        } ?: throw InvalidMIDIFile(path)

        val midi = try {
            Midi.from_bytes(bytes)
        } catch (e: Exception) {
            throw InvalidMIDIFile(path)
        }

        val opus_manager = this.get_opus_manager()
        opus_manager.project_change_midi(midi)
        val filename = this.parse_file_name(path.toUri())
        val new_path = this._project_manager.get_new_path()

        opus_manager.path = new_path
        opus_manager.set_project_name(filename?.substring(0, filename.lastIndexOf(".")) ?: getString(R.string.default_imported_midi_title))
        opus_manager.clear_history()
    }

    fun has_projects_saved(): Boolean {
        return this._project_manager.has_projects_saved()
    }

    fun is_soundfont_available(): Boolean {
        val soundfont_dir = this.get_soundfont_directory()
        return soundfont_dir.listFiles()?.isNotEmpty() == true
    }

    fun populate_supported_soundfont_instrument_names() {
        // populate a cache of available soundfont names so se don't have to open up the soundfont data
        // every time
        this._soundfont_supported_instrument_names.clear()
        val soundfont = this._soundfont
        if (soundfont != null) {
            for ((name, program, bank) in soundfont.get_available_presets()) {
                this._soundfont_supported_instrument_names[Pair(bank, program)] = name
            }
        } else {
            var program = 0
            for (name in this.resources.getStringArray(R.array.midi_instruments)) {
                this._soundfont_supported_instrument_names[Pair(0, program++)] = name
            }
        }
    }

    fun get_supported_instrument_names(): HashMap<Pair<Int, Int>, String> {
        if (this._soundfont_supported_instrument_names.isEmpty()) {
            this.populate_supported_soundfont_instrument_names()
        }

        return this._soundfont_supported_instrument_names
    }

    fun set_soundfont(filename: String?) {
        if (filename == null) {
            this.disable_soundfont()
            return
        }

        if (!this.update_playback_state_soundfont(PlaybackState.Ready)) {
            // Failed to change playback_state
            return
        }

        val path = "${this.getExternalFilesDir(null)}/SoundFonts/$filename"
        try {
            this._soundfont = SoundFont(path)
        } catch (e: Riff.InvalidRiff) {
            // Possible if user puts the sf2 in their files manually
            this.feedback_msg(getString(R.string.invalid_soundfont))
            return
        } catch (e: SoundFont.InvalidSoundFont) {
            // Possible if user puts the sf2 in their files manually
            this.feedback_msg("Invalid Soundfont")
            return
        }

        this.configuration.soundfont = filename
        this.populate_supported_soundfont_instrument_names()

        this.reinit_playback_device()
        this.connect_feedback_device()
        this.update_channel_instruments()
        this.populate_active_percussion_names()
        this.runOnUiThread {
            this.setup_project_config_drawer_export_button()
            val channel_recycler = this.findViewById<ChannelOptionRecycler>(R.id.rvActiveChannels)
            // Should always be null since this can only be changed from a different menu
            if (channel_recycler.adapter != null) {
                val channel_adapter = channel_recycler.adapter as ChannelOptionAdapter
                channel_adapter.notify_soundfont_changed()
            }
        }
    }

    fun update_playback_state_soundfont(next_state: PlaybackState): Boolean {
        this.playback_state_soundfont = this._update_playback_state(this.playback_state_soundfont, next_state) ?: return false
        return true
    }

    fun update_playback_state_midi(next_state: PlaybackState): Boolean {
        this.playback_state_midi = this._update_playback_state(this.playback_state_midi, next_state) ?: return false
        return true
    }

    private fun _update_playback_state(input_state: PlaybackState, next_state: PlaybackState): PlaybackState? {
        return when (input_state) {
            PlaybackState.NotReady -> {
                when (next_state) {
                    PlaybackState.NotReady,
                    PlaybackState.Ready -> next_state
                    else -> null
                }
            }
            PlaybackState.Ready -> {
                when (next_state) {
                    PlaybackState.NotReady,
                    PlaybackState.Ready,
                    PlaybackState.Queued -> next_state
                    else -> null
                }
            }
            PlaybackState.Playing -> {
                when (next_state) {
                    PlaybackState.Ready,
                    PlaybackState.Stopping -> next_state
                    else -> null
                }
            }
            PlaybackState.Queued -> {
                when (next_state) {
                    PlaybackState.Ready,
                    PlaybackState.Playing -> next_state
                    else -> null
                }
            }
            PlaybackState.Stopping -> {
                when (next_state) {
                    PlaybackState.Ready -> next_state
                    else -> null
                }
            }
        }
    }

    fun get_soundfont(): SoundFont? {
        return this._soundfont
    }

    fun disable_soundfont() {
        if (!this.update_playback_state_soundfont(PlaybackState.NotReady)) {
            return
        }

        if (this._feedback_sample_manager != null) {
            this.disconnect_feedback_device()
        }
        this._sample_handle_manager?.destroy()

        this._soundfont?.destroy()
        this._soundfont = null
        this._sample_handle_manager = null
        this.configuration.soundfont = null
        this._midi_playback_device = null
        this._soundfont_supported_instrument_names.clear()

        this.update_channel_instruments()
        this.populate_active_percussion_names()
    }

    fun get_soundfont_directory(): File {
        val soundfont_dir = File("${this.getExternalFilesDir(null)}/SoundFonts")
        if (!soundfont_dir.exists()) {
            soundfont_dir.mkdirs()
        }

        return soundfont_dir
    }

    fun save_configuration() {
        try {
            this.configuration.save(this._config_path)
        } catch (e: FileNotFoundException) {
            this.feedback_msg(resources.getString(R.string.config_file_not_found))
        }
    }

    fun get_drum_name(index: Int): String? {
        this.populate_active_percussion_names(false)
        return this.active_percussion_names[index + 27]
    }

    fun populate_active_percussion_names(force: Boolean = true) {
        if (force || this.active_percussion_names.isEmpty()) {
            this.active_percussion_names.clear()
            val drums = this.get_drum_options()
            for ((name, note) in drums) {
                if (note >= 27) {
                    this.active_percussion_names[note] = name
                }
            }
        }
    }

    fun parse_file_name(uri: Uri): String? {
        var result: String? = null
        if (uri.scheme == "content") {
            val cursor: Cursor? = contentResolver.query(uri, null, null, null, null)
            if (cursor != null) {
                if (cursor.moveToFirst()) {
                    val ci = cursor.getColumnIndex(OpenableColumns.DISPLAY_NAME)
                    if (ci >= 0) {
                        result = cursor.getString(ci)
                    }
                }
                cursor.close()
            }
        }

        if (result == null && uri.path is String) {
            result = uri.path!!
            result = result.substring(result.lastIndexOf("/") + 1)
        }
        return result
    }

    fun dialog_text_popup(title: String, default: String? = null, callback: (String) -> Unit) {
        val main_fragment = this.get_active_fragment()

        val viewInflated: View = LayoutInflater.from(main_fragment!!.context)
            .inflate(
                R.layout.text_input,
                main_fragment.view as ViewGroup,
                false
            )

        val text_input: EditText = viewInflated.findViewById(R.id.etText)
        text_input.setText(default ?: "")

        AlertDialog.Builder(this, R.style.Theme_Pagan_Dialog)
            .setTitle(title)
            .setView(viewInflated)
            .setPositiveButton(android.R.string.ok) { dialog, _ ->
                callback(text_input.text.toString())
                dialog.dismiss()
            }
            .setNeutralButton(android.R.string.cancel) { dialog, _ ->
                dialog.cancel()
            }
            .show()
    }

    fun dialog_name_and_notes_popup(default: Pair<String, String>? = null, callback: (String, String) -> Unit) {
        val main_fragment = this.get_active_fragment()

        val viewInflated: View = LayoutInflater.from(main_fragment!!.context)
            .inflate(
                R.layout.text_name_change,
                main_fragment.view as ViewGroup,
                false
            )

        val project_name_input: EditText = viewInflated.findViewById(R.id.etProjectName)
        project_name_input.setText(default?.first ?: "")

        val project_notes_input: EditText = viewInflated.findViewById(R.id.etProjectNotes)
        project_notes_input.setText(default?.second ?: "")

        AlertDialog.Builder(this, R.style.Theme_Pagan_Dialog)
            .setView(viewInflated)
            .setPositiveButton(android.R.string.ok) { dialog, _ ->
                callback(project_name_input.text.toString(), project_notes_input.text.toString())
                dialog.dismiss()
            }
            .setNeutralButton(android.R.string.cancel) { dialog, _ ->
                dialog.cancel()
            }
            .show()
    }

    internal fun dialog_popup_selection_offset() {
        val view_inflated: View = LayoutInflater.from(this)
            .inflate(
                R.layout.dialog_note_offset,
                window.decorView.rootView as ViewGroup,
                false
            )

        val np_octave = view_inflated.findViewById<NumberPicker>(R.id.npOctave)
        np_octave.maxValue = 14
        np_octave.minValue = 0
        np_octave.value = 7
        np_octave.wrapSelectorWheel = false
        np_octave.setFormatter { value: Int ->
            "${value - 7}"
        }

        val np_offset = view_inflated.findViewById<NumberPicker>(R.id.npOffset)
        val radix = (this.get_opus_manager().tuning_map.size - 1)
        np_offset.maxValue = (radix * 2)
        np_offset.minValue = 0
        np_offset.value = radix
        np_offset.wrapSelectorWheel = false
        np_offset.setFormatter { value: Int ->
            "${value - radix}"
        }

        this._popup_active = true
        AlertDialog.Builder(this, R.style.Theme_Pagan_Dialog)
            .setView(view_inflated)
            .setTitle(R.string.dialog_adjust_selection)
            .setOnDismissListener {
                this._popup_active = false
            }
            .setPositiveButton(android.R.string.ok) { dialog, _ ->
                val opus_manager = this.get_opus_manager()
                val radix = opus_manager.tuning_map.size
                val octave = np_octave.value - 7
                val offset = np_offset.value - (radix - 1)
                val real_delta = (octave * radix) + offset
                opus_manager.offset_selection(real_delta)

                this.get_action_interface().track(TrackedAction.AdjustSelection, listOf(real_delta))
                dialog.dismiss()
            }
            .setNeutralButton(android.R.string.cancel) { dialog, _ ->
                dialog.cancel()
            }
            .show()
    }

    internal fun <T> dialog_popup_menu(title: String, options: List<Pair<T, String>>, default: T? = null, callback: (index: Int, value: T) -> Unit) {
        if (this._popup_active) {
            return
        }

        if (options.isEmpty()) {
            return
        }

        this._popup_active = true
        val viewInflated: View = LayoutInflater.from(this)
            .inflate(
                R.layout.dialog_menu,
                window.decorView.rootView as ViewGroup,
                false
            )

        val recycler = viewInflated.findViewById<RecyclerView>(R.id.rvOptions)
        val dialog = AlertDialog.Builder(this, R.style.Theme_Pagan_Dialog)
            .setTitle(title)
            .setView(viewInflated)
            .setOnDismissListener {
                this._popup_active = false
            }
            .setNegativeButton(getString(android.R.string.cancel)) { dialog, _ ->
                dialog.dismiss()
            }
            .show()

        val adapter = PopupMenuRecyclerAdapter<T>(recycler, options, default) { index: Int, value: T ->
            dialog.dismiss()
            callback(index, value)
        }

        adapter.notifyDataSetChanged()
    }

    internal fun dialog_load_project() {
        val project_list = this._project_manager.get_project_list()

        this.dialog_popup_menu("Load Project", project_list) { _: Int, path: String ->
            this.get_action_interface().load_project(path)
        }
    }


    // TODO: fix code duplication in dialog_float/integer_input
    internal fun dialog_float_input(title: String, min_value: Float, max_value: Float, default: Float? = null, callback: (value: Float) -> Unit ) {
        val coerced_default_value = default ?: (this._float_dialog_defaults[title] ?: min_value)
        val viewInflated: View = LayoutInflater.from(this)
            .inflate(
                R.layout.dialog_float,
                window.decorView.rootView as ViewGroup,
                false
            )

        val number_input = viewInflated.findViewById<RangedFloatInput>(R.id.etNumber)

        val dialog = AlertDialog.Builder(this, R.style.Theme_Pagan_Dialog)
            .setCustomTitle(this._build_dialog_title_view(title))
            .setView(viewInflated)
            .setPositiveButton(android.R.string.ok) { _, _ ->
                val output_value = number_input.get_value() ?: coerced_default_value
                this._float_dialog_defaults[title] = output_value
                callback(output_value)
            }
            .setNeutralButton(android.R.string.cancel) { _, _ -> }
            .show()

        number_input.set_range(min_value, max_value)
        number_input.setText("$coerced_default_value")
        number_input.setOnClickListener {
            number_input.selectAll()
        }

        number_input.value_set_callback = { value: Float? ->
            callback(value ?: coerced_default_value)
            dialog.dismiss()
        }

        number_input.requestFocus()
        number_input.selectAll()
    }

    internal fun dialog_number_input(title: String, min_value: Int, max_value: Int, default: Int? = null, callback: (value: Int) -> Unit ) {
        val coerced_default_value = default ?: (this._integer_dialog_defaults[title] ?: min_value)
        val viewInflated: View = LayoutInflater.from(this)
            .inflate(
                R.layout.dialog_split,
                window.decorView.rootView as ViewGroup,
                false
            )

        val number_input = viewInflated.findViewById<RangedIntegerInput>(R.id.etNumber)

        val dialog = AlertDialog.Builder(this, R.style.Theme_Pagan_Dialog)
            .setTitle(title)
            .setView(viewInflated)
            .setPositiveButton(android.R.string.ok) { _, _ ->
                val output_value = number_input.get_value() ?: coerced_default_value
                this._integer_dialog_defaults[title] = output_value
                callback(output_value)
            }
            .setNeutralButton(android.R.string.cancel) { _, _ -> }
            .show()

        number_input.set_range(min_value, max_value)
        number_input.setText("$coerced_default_value")
        number_input.setOnClickListener {
            number_input.selectAll()
        }

        number_input.value_set_callback = { value: Int? ->
            callback(value ?: coerced_default_value)
            dialog.dismiss()
        }

        number_input.requestFocus()
        number_input.selectAll()
    }

    fun dialog_confirm(title: String, callback: () -> Unit) {
        AlertDialog.Builder(this, R.style.Theme_Pagan_Dialog)
            .setTitle(title)
            .setCancelable(true)
            .setPositiveButton(getString(R.string.dlg_confirm)) { dialog, _ ->
                dialog.dismiss()
                callback()
            }
            .setNegativeButton(getString(R.string.dlg_decline)) { dialog, _ ->
                dialog.dismiss()
            }
            .show()
    }

    private fun needs_save(): Boolean {
        val opus_manager = this.get_opus_manager()
        if (opus_manager.path == null) {
            return true
        }

        if (!File(opus_manager.path!!).exists()) {
            return !opus_manager.history_cache.isEmpty()
        }

        val other = OpusLayerBase()
        other.load_path(opus_manager.path!!)

        return opus_manager != other
    }

    fun dialog_save_project(callback: (Boolean) -> Unit) {
        if (this.needs_save()) {
            AlertDialog.Builder(this, R.style.Theme_Pagan_Dialog)
                .setTitle(R.string.dialog_save_warning_title)
                .setCancelable(true)
                .setPositiveButton(getString(R.string.dlg_confirm)) { dialog, _ ->
                    this@MainActivity.project_save()
                    dialog.dismiss()
                    callback(true)
                }
                .setNegativeButton(getString(R.string.dlg_decline)) { dialog, _ ->
                    dialog.dismiss()
                    callback(false)
                }
                .show()
        } else {
            callback(false)
        }
    }

    private fun dialog_delete_project() {
        val main_fragment = this.get_active_fragment()

        val title = this.get_opus_manager().project_name ?: getString(R.string.untitled_opus)

        AlertDialog.Builder(this, R.style.Theme_Pagan_Dialog)
            .setTitle(resources.getString(R.string.dlg_delete_title, title))
            .setPositiveButton(android.R.string.ok) { dialog, _ ->
                this@MainActivity.get_action_interface().delete()
                dialog.dismiss()
            }
            .setNegativeButton(android.R.string.cancel) { dialog, _ ->
                dialog.cancel()
            }
            .show()
    }

    fun select_import_file() {
        this.get_action_interface().import()
    }


    fun get_default_export_name(): String {
        val now = LocalDateTime.now()
        val formatter = DateTimeFormatter.ofPattern("yyyy-MM-dd")
        return "Pagan Op. ${now.format(formatter)}"
    }

    fun get_export_name(): String {
        val reserved_chars = "|\\?*<\":>+[]/'"
        var base_name: String = this.get_opus_manager().project_name ?: this.get_default_export_name()
        for (c in reserved_chars) {
            base_name = base_name.replace("$c", "_")
        }
        return base_name
    }

    fun export_multi_lines_wav() {
        val name = this.get_export_name()
        val intent = Intent(Intent.ACTION_OPEN_DOCUMENT_TREE)
        intent.putExtra(Intent.EXTRA_TITLE, name)
        this._export_multi_line_wav_intent_launcher.launch(intent)
    }

    fun export_multi_channels_wav() {
        val name = this.get_export_name()
        val intent = Intent(Intent.ACTION_OPEN_DOCUMENT_TREE)
        intent.putExtra(Intent.EXTRA_TITLE, name)
        this._export_multi_channel_wav_intent_launcher.launch(intent)
    }

    fun export_wav() {
        val name = this.get_export_name()
        val intent = Intent(Intent.ACTION_CREATE_DOCUMENT)
        intent.addCategory(Intent.CATEGORY_OPENABLE)
        intent.type = "audio/wav"
        intent.putExtra(Intent.EXTRA_TITLE, "$name.wav")
        this._export_wav_intent_launcher.launch(intent)
    }

    fun export_wav_cancel() {
        this.view_model.cancel_export()
    }

    fun export_midi() {
        val name = this.get_export_name()

        val intent = Intent(Intent.ACTION_CREATE_DOCUMENT)
        intent.addCategory(Intent.CATEGORY_OPENABLE)
        //intent.type = MimeTypes.AUDIO_MIDI
        intent.type = "audio/midi"
        intent.putExtra(Intent.EXTRA_TITLE, "$name.mid")

        this._export_midi_intent_launcher.launch(intent)
    }

    fun export_project() {
        val name = this.get_export_name()

        val intent = Intent(Intent.ACTION_CREATE_DOCUMENT)
        intent.addCategory(Intent.CATEGORY_OPENABLE)
        intent.type = "application/json"
        intent.putExtra(Intent.EXTRA_TITLE, "$name.json")

        this._export_project_intent_launcher.launch(intent)
    }

    /**
     * Save text file in storage of a crash report.
     * To be copied and saved somewhere accessible on reload.
     */
    fun bkp_crash_report(e: Throwable) {
        val path = this.getExternalFilesDir(null).toString()
        val file = File("$path/bkp_crashreport.log")
        file.writeText(e.stackTraceToString())
    }

    fun check_for_crash_report() {
        val path = this.getExternalFilesDir(null).toString()
        val file = File("$path/bkp_crashreport.log")
        if (file.isFile) {
            AlertDialog.Builder(this, R.style.Theme_Pagan_Dialog)
               .setTitle(R.string.crash_report_save)
                .setMessage(R.string.crash_report_desc)
                .setCancelable(true)
                .setPositiveButton(getString(R.string.dlg_confirm)) { dialog, _ ->
                    export_crash_report()
                    dialog.dismiss()
                }
                .setNegativeButton(getString(R.string.dlg_decline)) { dialog, _ ->
                    file.delete()
                    dialog.dismiss()
                }
                .show()
        }
    }

    fun export_crash_report() {
        val now = LocalDateTime.now()
        val formatter = DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm")
        val name = "pagan.cr-${now.format(formatter)}.log"

        val intent = Intent(Intent.ACTION_CREATE_DOCUMENT)
        intent.addCategory(Intent.CATEGORY_OPENABLE)
        //intent.type = MimeTypes.AUDIO_MIDI
        intent.type = "text/plain"
        intent.putExtra(Intent.EXTRA_TITLE, "$name")

        this._crash_report_intent_launcher.launch(intent)
    }

    fun set_sample_rate(new_sample_rate: Int) {
        this.configuration.sample_rate = new_sample_rate
        this.save_configuration()
        this.set_soundfont(this.configuration.soundfont)
    }

    fun reinit_playback_device() {
        this._midi_playback_device?.kill()

        if (this.get_soundfont() != null) {
            /*
             * TODO: Put the ignore envelope/lfo option somewhere better.
             * I don't think it should be in apres if theres a reasonable way to avoid it
             */
            this._sample_handle_manager?.destroy()
            this._sample_handle_manager = SampleHandleManager(
                this._soundfont!!,
                this.configuration.sample_rate,
                this.configuration.sample_rate,
                ignore_lfo = true
            )

            this._midi_playback_device = PlaybackDevice(
                this,
                this._sample_handle_manager!!,
                WaveGenerator.StereoMode.Stereo
            )
        } else {
            this._midi_playback_device = null
        }
    }

    fun in_playback(): Boolean {
        return PlaybackState.Playing in listOf(this.playback_state_soundfont, this.playback_state_midi)
    }

    fun has_notification_permission(): Boolean {
        return (Build.VERSION.SDK_INT < Build.VERSION_CODES.TIRAMISU || ActivityCompat.checkSelfPermission(this, android.Manifest.permission.POST_NOTIFICATIONS) == PackageManager.PERMISSION_GRANTED )
    }

    private fun getNotificationPermission(): Boolean {
        if (! this.has_notification_permission()) {
            ActivityCompat.requestPermissions(this, arrayOf(android.Manifest.permission.POST_NOTIFICATIONS), 100)
        }

        return this.has_notification_permission()
    }

    fun set_playback_button(drawable: Int) {
        val play_pause_button = this._options_menu?.findItem(R.id.itmPlay) ?: return
        play_pause_button.icon = ContextCompat.getDrawable(this, drawable)
    }

    fun set_midi_playback_button(drawable: Int) {
        val play_pause_button = this._options_menu?.findItem(R.id.itmPlayMidiOutput) ?: return
        play_pause_button.icon = ContextCompat.getDrawable(this, drawable)
    }

    fun get_working_column(): Int {
        val cursor = this.get_opus_manager().cursor
        return when (cursor.mode) {
            OpusManagerCursor.CursorMode.Single,
            OpusManagerCursor.CursorMode.Column -> {
                cursor.beat
            }
            OpusManagerCursor.CursorMode.Range -> {
                cursor.get_ordered_range()!!.first.beat
            }
            else -> {
                val editor_table = this.findViewById<EditorTable?>(R.id.etEditorTable)
                editor_table?.get_first_visible_column_index() ?: 0
            }
        }
    }

    fun is_connected_to_physical_device(): Boolean {
        return this.configuration.allow_midi_playback && this._midi_interface.output_devices_connected()
    }

    fun disconnect_feedback_device() {
        this._temporary_feedback_devices.forEachIndexed { i: Int, device: FeedbackDevice? ->
            device?.kill()
            device?.destroy()
            this._temporary_feedback_devices[i] = null
        }
        this._feedback_sample_manager?.destroy()
        this._feedback_sample_manager = null
    }

    fun connect_feedback_device() {
        if (this._soundfont == null) {
            return
        }

        this.disconnect_feedback_device()

        val buffer_size = this.configuration.sample_rate / 2
        this._feedback_sample_manager = SampleHandleManager(
            this._soundfont!!,
            this.configuration.sample_rate,
            buffer_size - 2 + (if (buffer_size % 2 == 0) { 2 } else { 1 })
        )
        this._current_feedback_device = 0
    }

    fun block_physical_midi_output() {
        this._midi_interface.block_physical_devices = true
        this._midi_interface.close_connected_devices()
        this.playback_state_midi = PlaybackState.NotReady

        if (this._feedback_sample_manager == null) {
            this.connect_feedback_device()
        }
    }

    fun enable_physical_midi_output() {
        this._midi_interface.block_physical_devices = false
        this._midi_interface.open_connected_devices()
        this.playback_state_midi = PlaybackState.Ready

        if (this.is_connected_to_physical_device()) {
            this.disconnect_feedback_device()
        }
    }

    fun get_notification(): NotificationCompat.Builder? {
        if (!this.has_notification_permission()) {
            return null
        }

        if (this._active_notification == null) {
            this.get_notification_channel()

            val cancel_export_flag = "com.qfs.pagan.CANCEL_EXPORT_WAV"
            val intent = Intent()
            intent.setAction(cancel_export_flag)
            intent.setPackage(this.packageName)

            val pending_cancel_intent = PendingIntent.getBroadcast(
                this,
                1,
                intent,
                PendingIntent.FLAG_IMMUTABLE
            )

            val builder = NotificationCompat.Builder(this, CHANNEL_ID)
                .setContentTitle(this.getString(R.string.export_wav_notification_title, this.get_opus_manager().project_name ?: "Untitled Project"))
                .setPriority(NotificationCompat.PRIORITY_LOW)
                .setSmallIcon(R.drawable.small_logo_rowan)
                .setSilent(true)
                .addAction(R.drawable.baseline_cancel_24, this.getString(android.R.string.cancel), pending_cancel_intent)

            this._active_notification = builder
        }

        return this._active_notification!!
    }

    fun get_notification_channel(): NotificationChannel? {
        return if (!this.has_notification_permission()) {
            null
        } else if (this._notification_channel == null) {
            val notification_manager = NotificationManagerCompat.from(this)
            // Create the NotificationChannel.
            val name = this.getString(R.string.export_wav_file_progress)
            val descriptionText = this.getString(R.string.export_wav_notification_description)
            val importance = NotificationManager.IMPORTANCE_DEFAULT
            val mChannel = NotificationChannel(CHANNEL_ID, name, importance)
            mChannel.description = descriptionText
            // Register the channel with the system. You can't change the importance
            // or other notification behaviors after this.
            notification_manager.createNotificationChannel(mChannel)
            mChannel
        } else {
            this._notification_channel!!
        }
    }

    private fun _refresh_toolbar() {
        val toolbar = this._binding.appBarMain.toolbar

        when (this.get_active_fragment()) {
            is FragmentEditor -> {
                toolbar.setNavigationIcon(R.drawable.hamburger_32)
            }
            is FragmentLandingPage -> {
                toolbar.navigationIcon = null
            }
            else -> {
                toolbar.setNavigationIcon(R.drawable.baseline_arrow_back_24)
            }
        }
    }

    fun vibrate() {
        val vibrator = if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.S) {
            val vibratorManager = this.getSystemService(Context.VIBRATOR_MANAGER_SERVICE) as VibratorManager
            vibratorManager.defaultVibrator
        } else {
            @Suppress("DEPRECATION")
            getSystemService(Context.VIBRATOR_SERVICE) as Vibrator
        }

        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.Q) {
            vibrator.vibrate(VibrationEffect.createOneShot(200, VibrationEffect.EFFECT_HEAVY_CLICK))
        } else {
            vibrator.vibrate(VibrationEffect.createOneShot(200, 5))
        }
    }


    fun get_file_type(path: String): CompatibleFileType {
        return this.applicationContext.contentResolver.openFileDescriptor(Uri.parse(path), "r")?.use {
            val test_bytes = ByteArray(4)
            FileInputStream(it.fileDescriptor).read(test_bytes)
            if (test_bytes.contentEquals("MThd".toByteArray())) {
                CompatibleFileType.Midi1
            } else {

                CompatibleFileType.Pagan
            }
        } ?: throw FileNotFoundException(path)
    }

    fun get_action_interface(): ActionTracker {
        return this.view_model.action_interface
    }

    fun save_actions() {
        val generated_code = this.get_action_interface().to_json().to_string()
        val path = this.getExternalFilesDir(null).toString()
        val timestamp = System.currentTimeMillis()
        val file_name = "$path/generated_$timestamp.json"
        val file = File(file_name)
        file.writeText(generated_code)
        this.get_action_interface().clear()
    }

    fun navigate_import(uri: Uri) {
        val fragment = this.get_active_fragment()
        fragment?.clearFragmentResult(IntentFragmentToken.Resume.name)
        fragment?.setFragmentResult(
            IntentFragmentToken.ImportGeneral.name,
            bundleOf(Pair("URI", uri.toString()))
        )
        if (fragment !is FragmentEditor) {
            this.navigate(R.id.EditorFragment)
        }
    }

    fun is_debug_on(): Boolean {
        return this.packageName.contains("pagandev")
    }

    override fun onDestroy() {
        this._sample_handle_manager?.destroy()
        this._feedback_sample_manager?.destroy()
        super.onDestroy()

    }

    fun set_forced_orientation(value: Int) {
        this.configuration.force_orientation = value
        this.save_configuration()
        this.requestedOrientation = value
    }

}
